import java.util.*;
import java.io.*;
import java.awt.Color;
import org.xml.sax.helpers.*;
import org.xml.sax.*;
import javax.xml.parsers.*;

// WORK IN PROGESS

/** 
 * Example of Java classes for astronomy in Sundog Resurection
 * <P>
 * This models the original game Sundog Frozen Legacy.
 * <P>
 * Sector is the main class (root class of the tree)
 * <BR>A sector is a set of star systems.
 *
 * @author Arnaud Cheritat
 */
class Sector
{
    /**
     * The set of systems in the sector.<BR>
     * Private: user of this class has no direct access.
     * Access only through Sector.add(StarSystem s) and Sector.remove(...).
     * This prevents user from adding anything else than systems.
     */
    private HashSet systems;

    /** The name of the sector. */
    public String name;

    /** Comments */
    public String comment;

    /** Used to tell if last loading has been successful */
    private boolean loadSuccess;

    /** Creates an empty sector */
    public Sector() {
	systems = new HashSet();
	name = "";
	loadSuccess=false;
	comment=null;
    }

    /** Add a system */
    public void add(StarSystem s) {
	systems.add(s);
    }

    /** Remove a system */
    public void remove(StarSystem s) {
	systems.remove(s);
    }

    /** Save the Sector to an XML file*/
    public void save(File file) {
	try {	    
	    FileOutputStream fileos=new FileOutputStream(file);

	    // OutputStreamWriter converts 16-bits chars and strings of Java
	    // into bytes, according to some encoding scheme
	    OutputStreamWriter osw=new OutputStreamWriter(fileos);

	    // The following is supposed to increase efficiency
	    // Don't know if it is necessary
	    BufferedWriter f=new BufferedWriter(osw);

	    // XML declaration
	    f.write("<?xml version=\"1.0\" encoding=\"");
	    //	    f.write(osw.getEncoding());
	    f.write("iso-8859-1");
	    f.write("\"?>\n");

	    // Any DTD should be declared here
	    f.write("<!DOCTYPE sector SYSTEM \"v6.dtd\">\n");
	    /// no DTD	   

	    // Some comment
	    f.write("<!-- Sector description for 'Sundog Resurection' -->\n");

	    //
	    f.write("<sector>\n");

	    if(comment!=null) {
		f.write("<comment>");
		f.write(comment);
		f.write("</comment>");
	    }

	    Iterator systemI = iterator();
	    for(; systemI.hasNext(); ) {
		StarSystem s=(StarSystem) systemI.next();
		if(s.spec==StarSystem.REGULAR) {
		    f.write(" <system>\n");
		}
		else {
		    f.write(" <system spec=\"");
		    f.write(s.spec==StarSystem.BEGINNING ? "beginning" :
			    s.spec==StarSystem.UNREACHABLE ? "unreachable" :
			    "ERROR");
		    f.write("\">");
		}

		f.write("  <name>");
		f.write(s.name);
		f.write("</name>\n");

		if(s.mainStar!=null) {
		    f.write("  <star name=\"");
		    f.write(s.mainStar.name);
		    f.write("\" class=\"");
		    f.write(s.mainStar.starClass);
		    f.write("\" mass=\"");
		    f.write(String.valueOf(s.mainStar.mass));
		    f.write("\" diameter=\"");
		    f.write(String.valueOf(s.mainStar.diameter));
		    f.write("\"/>\n");
		}

		f.write("  <coords x=\"");
		f.write(String.valueOf(s.x));
		f.write("\" y=\"");
		f.write(String.valueOf(s.y));
		f.write("\" z=\"");
		f.write(String.valueOf(s.z));
		f.write("\"/>\n");

		if(s.comment!=null) {
		    f.write("<comment>");
		    f.write(s.comment);
		    f.write("</comment>");
		}

		Iterator planetI = s.iterator();
		for(; planetI.hasNext(); ) {
		    Planet p=(Planet) planetI.next();
		    f.write("  <planet>\n");

		    f.write("   <name>");
		    f.write(p.name);
		    f.write("</name>\n");

		    f.write("   <orbit radius=\"");
		    f.write(String.valueOf(p.orbitRadius));
		    f.write("\" initAngle=\"");
		    f.write(String.valueOf(p.orbitInitAngle));
		    f.write("\"/>");

		    /*
		    f.write("   <land type=\"");		    
		    f.write(p.landType==Planet.PLAIN ? "plain" :
			    p.landType==Planet.JUNGLE ? "jungle" :
			    p.landType==Planet.DESERT ? "desert" :
			    p.landType==Planet.ICE ? "ice" :
			    p.landType==Planet.CRATERED ? "cratered" :
			    "ERROR");
		    f.write("\"/>");
		    */

		    if(p.comment!=null) {
			f.write("<comment>");
			f.write(p.comment);
			f.write("</comment>");
		    }

		    Iterator cityI = p.iterator();
		    for(; cityI.hasNext(); ) {
			City c=(City) cityI.next();
			
			if(c.type==City.REGULAR) {
			    f.write("   <city>\n");
			}
			else {
			    f.write("   <city spec=\"");
			    f.write(c.type==City.STARPORT ? "Starport" :
				    c.type==City.COLONY ? "Colony" :
				    c.type==City.NO_LANDING ? "NoLanding" :
				    "ERROR");
			    f.write("\">");
			}

			f.write(c.name);
			
			f.write("   </city>\n");
		    }

		    f.write("  </planet>\n");
		}
		
		f.write(" </system>\n");
	    }

	    f.write("</sector>\n");
	    f.close();
	}
	catch(Exception e) {
	    exceptionMessage(e);
	}
    }

    /**
     * Load Sector from an XML file<br>
     * CAUTION: reading destroys the old sector, even if a read error
     * happens. Programmer using this class must handle this by
     * backing-up the old sector.
     */
    public boolean load(File file) {
	try {
	    //FileInputStream fileos=new FileInputStream(file);
	    //InputStreamReader osr=new InputStreamReader(fileos);
	    //BufferedReader fileReader=new BufferedReader(osr);
	    	   
	    DefaultHandler handler = new MyParser();
	    SAXParserFactory factory = SAXParserFactory.newInstance();
	    factory.setValidating(true);
	    SAXParser saxParser = factory.newSAXParser();
	    systems=new HashSet();
	    saxParser.parse( file, handler );
	    return(loadSuccess);
	}
	catch(Exception e) {
	    exceptionMessage(e);
	    return(false);
	}
    }

    public Iterator iterator() {
	return(systems.iterator());
    }

    /**
     * Handles exceptions messages.
     */
    public void exceptionMessage(Exception e) {
	System.out.println(e.getMessage());
    }

    /**
     * To use the java implementation of the XML reading tool named SAX
     * one way is to subclass the DefaultHandler class.
     */
    private class MyParser extends DefaultHandler
    {
	private static final int NONE=0, SECTOR=1, SYSTEM=2,
	    PLANET=3, CITY=4;
	private int level;
	private static final int NAME=101,COMMENT=102; 
	private int reading;

	// I could have chosen "private Object currentObject;"
	// but porting would be harder
	private StarSystem currentStarSystem;
	private Planet currentPlanet;
	private City currentCity;

	private String pcData;

	private boolean problem,readStar;
	
	void MyParser() {
	    level=NONE;
	    reading=NONE;
	    currentPlanet=null;
	    currentStarSystem=null;
	    currentCity=null;
	    pcData=""; // facultative ?

	    loadSuccess=false; // this variable belongs to the parent
	    problem=false;
	    readStar=false; // maybe redundant
	}

	public void warning(SAXParseException e)
	    throws SAXException	{
	    System.out.println("XML warning : "+e.getMessage());
	}

	public void error(SAXParseException e)
	    throws SAXException	{
	    System.out.println("XML recoverable error : "+e.getMessage());
	    problem=true;
	}
	public void fatalError(SAXParseException e)
	    throws SAXException	{
	    System.out.println("XML fatal error : "+e.getMessage());
	    problem=true;
	}

	public void startElement(String namespaceURI,
                         String localName,
                         String qName,
                         Attributes atts)
	    throws SAXException
	{

	    reading=NONE;
	    if(qName.equals("name")) {
		reading=NAME;
		pcData="";
	    }
	    if(qName.equals("comment")) {
		reading=COMMENT;
		pcData="";
	    }
	    if(qName.equals("coords")) {
		if(currentStarSystem!=null && level==SYSTEM) {
		    try {
			currentStarSystem.x=Double.valueOf(atts.getValue("x")).doubleValue();
			currentStarSystem.y=Double.valueOf(atts.getValue("y")).doubleValue();
			currentStarSystem.z=Double.valueOf(atts.getValue("z")).doubleValue();
		    } catch(NumberFormatException e) {
			throw(new SAXException("Error while reading coordinates"));
		    } catch(NullPointerException e) {
			problem=true;
			throw(e);
		    }
		}
		else {
		    problem=true;
		    throw(new SAXException("Error : attempting to assign coordinates at wrong place"));
		}
	    }

	    if(qName.equals("star")) {	       
		if(level==SYSTEM && currentStarSystem!=null) {
		    if(!readStar) {
			try {
			    currentStarSystem.mainStar.name=atts.getValue("name");
			    currentStarSystem.mainStar.starClass=atts.getValue("class");
			    currentStarSystem.mainStar.mass=Double.valueOf(atts.getValue("mass")).doubleValue();
			    currentStarSystem.mainStar.diameter=Double.valueOf(atts.getValue("diameter")).doubleValue();
			    double lum=Double.valueOf(atts.getValue("luminosity")).doubleValue();
			    if(lum<0.00001) {
				currentStarSystem.mainStar.magnitude=13.0;
			    } else {
				currentStarSystem.mainStar.magnitude=-2.5*Math.log(lum);
			    }
			    readStar=true;
			} catch(NumberFormatException e) {
			    throw(new SAXException("Error while reading star properties"));
			} catch(NullPointerException e) {
			    problem=true;
			    throw(e);
			}
		    }
		}
		else {
		    problem=true;
		    throw(new SAXException("Error : trying to set set star properties at wrong level"));
		}
	    }
	    if(qName.equals("sector")){
		level=SECTOR;
	    }
	    if(qName.equals("system")) {
		readStar=false; // maybe redundant
		level=SYSTEM;
		currentStarSystem=new StarSystem();
		systems.add(currentStarSystem);
		String spec=atts.getValue("spec");
		if(spec!=null) {
		    currentStarSystem.spec =
			spec.equals("beginning") ? StarSystem.BEGINNING
		      : spec.equals("unreachable") ? StarSystem.UNREACHABLE
		      : StarSystem.REGULAR;
		}
		else {
		    currentStarSystem.spec = StarSystem.REGULAR;
		}
	    }
	    if(qName.equals("planet")){
		level=PLANET;
		currentPlanet=new Planet();
		if(currentStarSystem!=null) {
		    currentStarSystem.add(currentPlanet);
		}
		else {
		    problem=true;
		    throw(new SAXException("Error : adding a planet whereas no star system is defined"));
		}
	    }
	    if(qName.equals("orbit")){
		if(level==PLANET && currentPlanet!=null) {
		    try {
			currentPlanet.orbitRadius=Double.valueOf(atts.getValue("radius")).doubleValue();
			currentPlanet.orbitInitAngle=Double.valueOf(atts.getValue("initAngle")).doubleValue();
		    } catch(NumberFormatException e) {
			throw(new SAXException("Error while reading planet orbit"));
		    } catch(NullPointerException e) {
			problem=true;
			throw(e);
		    }
		}
		else {
		    problem=true;
		    throw(new SAXException("Error : setting planet orbit at wrong place"));
		}
	    }
	    /*
	    if(qName.equals("land")){
		if(currentPlanet!=null && level==PLANET) {
		    String str=atts.getValue("type");
		    currentPlanet.landType=Planet.NONE;
		    if(str.equals("jungle")) {
			currentPlanet.landType=Planet.JUNGLE;
		    }
		    if(str.equals("plain")) {
			currentPlanet.landType=Planet.PLAIN;
		    }
		    if(str.equals("desert")) {
			currentPlanet.landType=Planet.DESERT;
		    }
		    if(str.equals("jungle")) {
			currentPlanet.landType=Planet.ICE;
		    }
		    if(str.equals("cratered")) {
			currentPlanet.landType=Planet.CRATERED;
		    }
		    if(currentPlanet.landType==NONE) {
			throw(new SAXException("Error : unknown land type"));
		    }
		}
		else {
		    throw(new SAXException("Error : setting land at wrong place"));
		}
	    }
	    */
	    if(qName.equals("city")){
		level=CITY;
		currentCity=new City();
		if(currentPlanet!=null) {
		    currentPlanet.add(currentCity);
		}
		else {
		    problem=true;
		    throw(new SAXException("Error : adding a city whereas no planet is defined"));
		}
		reading=CITY;
		pcData="";
	    }
	}

	public void endElement(String uri,
			       String localName,
			       String qName)
	    throws SAXException
	{
	    if(qName.equals("name")) {
		switch(level) {
		case SECTOR :
		    name=new String(pcData);
		    break;
		case SYSTEM :
		    if(currentStarSystem!=null) {
			currentStarSystem.name=new String(pcData);
		    }
		    else {
			problem=true;
			throw(new SAXException("Error, trying to assign name at wrong place (1)"));
		    }
		    break;
		case PLANET :
		    if(currentPlanet!=null) {
			currentPlanet.name=new String(pcData);
		    }
		    else {
			problem=true;
			throw(new SAXException("Error, trying to assign name at wrong place (2)"));
		    }
		    break;
		default :
		    problem=true;
		    throw(new SAXException("Error, trying to assign name at wrong place (3)"));
		}
		reading=NONE;
	    }

	    if(qName.equals("comment")) {
		switch(level) {
		case SECTOR :
		    comment=new String(pcData);
		    break;
		case SYSTEM :
		    if(currentStarSystem!=null) {
			currentStarSystem.comment=new String(pcData);
		    }
		    else {
			problem=true;
			throw(new SAXException("Error, trying to set comment at wrong place (1)"));
		    }
		    break;
		case PLANET :
		    if(currentPlanet!=null) {
			currentPlanet.comment=new String(pcData);
		    }
		    else {
			problem=true;
			throw(new SAXException("Error, trying to set comment at wrong place (2)"));
		    }
		    break;
		default :
		    problem=true;
		    throw(new SAXException("Error, trying to set comment at wrong place (3)"));
		}
		reading=NONE;
	    }

	    if(qName.equals("city")) {
		if(currentCity!=null & level==CITY) {
			currentCity.name=new String(pcData);
		}
		else {
		    problem=true;
		    throw(new SAXException("Error, trying to name city at wrong place"));
		}
		level=PLANET;
		reading=NONE;
	    }

	    if(qName.equals("sector")) {
		level=NONE;
	    }

	    if(qName.equals("system")) {
		level=SECTOR;
		readStar=false; // maybe redundant
	    }

	    if(qName.equals("planet")) {
		level=SYSTEM;
	    }
	}

	public void characters(char[] ch,
			       int start,
			       int length)
	    throws SAXException {
	    //pcData.append(new String(ch,start,length)));
	    if(reading!=NONE) {
		pcData=pcData.concat(new String(ch,start,length));
	    }
	}

	public void ignorableWhitespace(char[] ch,
					int start,
					int length)
	    throws SAXException
	{
	    /*
	    if(reading!=NONE && pcData!=null) {
		pcData=pcData.concat(" ");
	    }
	    */
	}

	public void endDocument()
	    throws SAXException {
	    loadSuccess=!problem;
	}

    }
}

class Star
{
    /*
    public static final int
	WHITE=1,
	BRIGHT_BLUE=2,
	BLUE=3,
	BRIGHT_YELLOW=4,
	YELLOW=5,
	ORANGE=6,
	RED=7,
	DARK_RED=8;
    */

    public String name;
    public String starClass;
    public double mass;
    public double diameter;
    public double luminosity;
    public double magnitude;

    public Star() {
	name="";
	starClass="G2";
	mass=0;
	diameter=0;
	luminosity=0;
    }

    public Color getColor() {
	Color color;
	if(starClass!=null) {
	    if(starClass.length()>0) {
		char c=starClass.charAt(0);
		switch(c) {
		    /*
		      O     155 176 255  #9bb0ff  
		      B     170 191 255  #aabfff  
		      A     202 215 255  #cad7ff  
		      F     248 247 255  #f8f7ff  
		      G     255 244 234  #fff4ea  
		      K     255 210 161  #ffd2a1  
		      M     255 204 111  #ffcc6f  
		    */
		    /*
		case 'O' : color=new Color(155,176,255); break;
		case 'B' : color=new Color(170,191,255); break;
		case 'A' : color=new Color(202,215,255); break;
		case 'F' : color=new Color(248,247,255); break;
		case 'G' : color=new Color(255,244,234); break;
		case 'K' : color=new Color(255,210,161); break;
		case 'M' : color=new Color(255,204,111); break;
		case 'E' : color=new Color(255,82,0); break;
		default : color=new Color(255,244,234);
		    */
		case 'O' : color=new Color(0.0f,0.0f,1.0f); break; // bluer
		case 'B' : color=new Color(0.0f,0.5f,1.0f); break; // blue
		case 'A' : color=new Color(0.0f,0.9f,0.9f); break; // cyan
		case 'F' : color=new Color(0.9f,0.9f,0.9f); break; // White
		case 'G' : color=Color.YELLOW; break;
		case 'K' : color=Color.ORANGE; break;
		case 'M' : color=Color.RED; break;
		case 'E' : color=new Color(0.7f,0,0); break;
		default : color=Color.YELLOW;
		    /*
		case 'O' : color=WHITE; break;
		case 'B' : color=BRIGHT_BLUE; break;
		case 'A' : color=BLUE; break;
		case 'F' : color=BRIGHT_YELLOW; break;
		case 'G' : color=YELLOW; break;
		case 'K' : color=ORANGE; break;
		case 'M' : color=RED; break;
		case 'E' : color=DARK_RED; break;
		    */
		}
	    }
	    else {
		color=Color.YELLOW;
	    }
	}
	else {
	    color=Color.YELLOW;
	}
	return(color);
    }
}

class StarSystem
{
    /** Identifier for usual systems */
    public static final int REGULAR=0;

    /** Identifier for the beginning system (Jondd in the original game) */
    public static final int BEGINNING=1;

    /** Identifier for system too far (Enlie in the original game) */
    public static final int UNREACHABLE=2;

    /** The set of planets */
    private HashSet planets;

    /** System's Name (usually, name of the main star)*/
    public String name;

    /** Comments */
    public String comment;

    public int spec;
    public double x,y,z;

    public Star mainStar;


    /** The name of the system. */
    public StarSystem() {
	planets = new HashSet();
	name = "";
	spec=REGULAR;
	x=0;
	y=0;
	z=0;
	comment=null;
	mainStar=new Star();
    }

    public void add(Planet p) {
	planets.add(p);
    }

    public void remove(Planet p) {
	planets.remove(p);
    }

    public Iterator iterator() {
	return(planets.iterator());
    }
}

class Planet
{
    /*
    // Should the land type depend on the distance to the sun ?

    public static final int NONE=0;
    public static final int PLAIN=1;
    public static final int JUNGLE=2;
    public static final int DESERT=3;
    public static final int ICE=4;
    public static final int CRATERED=1;
    */

    /** The set of cities on the planet. */
    private HashSet cities;

    /** The name of the planet. */
    public String name;

    /** Comments */
    public String comment;

    /** The radius of the orbit*/
    public double orbitRadius;
    public double orbitInitAngle;

    /**
     * The type of land (plain, jungle, desert, etc...).
     * Maybe it would be better to let the type depend on the distance
     * to the star.
     */
//    public int landType;

    // There remains to put orbital parameters

    public Planet() {
	cities = new HashSet();
	name = "";
//	landType=PLAIN;
	comment=null;
    }

    public void add(City c) {
	cities.add(c);
    }

    public void remove(City c) {
	cities.remove(c);
    }

    public Iterator iterator() {
	return(cities.iterator());
    }
}

class City {
    /** Identifier for regular cities */
    public static final int REGULAR=0;

    /** Identifier for Starports */
    public static final int STARPORT=1;

    /** Identifier for the Colony (Banville in the original game) */
    public static final int COLONY=2;

    /**
     * Identifier for the city in which one can not land
     * (I think it was Tuie, in Enlie system in the original game)
     */
    public static final int NO_LANDING=3;

    /** Name of the city */
    public String name;

    /** Comments */
    public String comment;

    /** Type of the city (REGULAR, STARPORT, etc...) */
    public int type;

    public City() {
	type = REGULAR;
	name = "";
	comment=null;
    }
}
